@@ -315,7 +315,10 @@ EXFUN (Fextent_start_position, 1);
 EXFUN (Fmake_extent, 3);
 EXFUN (Fprevious_single_property_change, 4);
 EXFUN (Fset_extent_endpoints, 4);
+EXFUN (Fnext_extent_change, 2);
+EXFUN (Fprevious_extent_change, 2);
 EXFUN (Fset_extent_parent, 2);
+EXFUN (Fget_char_property, 4);
 
 extern int inside_undo;
 extern int in_modeline_generation;
@@ -325,7 +328,7 @@ struct extent_fragment *extent_fragment_new (Lisp_Object buffer_or_string,
 face_index extent_fragment_update (struct window *w,
 				   struct extent_fragment *ef,
 				   /* Note this is in Bytinds */
-				   Bytind pos);
+				   Bytind pos, Lisp_Object last_glyph);
 void extent_fragment_delete (struct extent_fragment *ef);
 
 
